using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using Microsoft.SPOT.Hardware.SJJ;
using Microsoft.SPOT.Hardware.SJJ.SPI_LCD_Driver;
using System.Threading;

namespace CombinationLock
{
    public class Program
    {
        public static void Main()
        {
            Debug.Print(Resources.GetString(Resources.StringResources.String1));
            Debug.Print(Resources.GetString(Resources.StringResources.String2));
            App myApp = new App();
            myApp.Run();
        }

    }

    public class App
    {
        DispScreen myDispScreen = new DispScreen();
        static LockStateMachine.eLockState eCurrentLockState;
        static LockAccount myLockAccount = new LockAccount();
        const Boolean bNoEcho = false;
        const Boolean bEcho = true;
        const Boolean bUnlock = true;
        const Boolean bLock = false;
        static int iSaveRetries;
        static int iSaveLockOut;
        public enum eLogViewState
        {
            DontResetCounts,
            ResetCounts,
            CancelLogView
        }

        public App()
        {
            eCurrentLockState = LockStateMachine.eLockState.MainState;
        }

        public void Run()
        {
            LockStateMachine  myLockState = new LockStateMachine();
            UInt32 uiLockCode = 0;
            OutputPort CPLDX0 = new OutputPort(Pins.GPIO_PORT_X_1, bLock);

            if (!myLockAccount.GetData())
            {
                myLockAccount.SaveData();
            }

            while (true)
            {
                eCurrentLockState = myLockState.GetLockState();
                switch (eCurrentLockState)
                {
                    case LockStateMachine.eLockState.MainState:
                        {
                            myDispScreen.ShowMainScreen();

                            //Get keypad input
                            uiLockCode = myDispScreen.GetLockCode(bNoEcho);

                            // If lock code valid user code, set state to UnlockState
                            switch (myLockAccount.CheckLockCode(uiLockCode))
                            {
                                case LockAccount.eLockCodeType.InvalidCode:
                                    {
                                        myLockAccount.IncrementFailedCodeCount();

                                        if (myLockAccount.CheckCodeFailures())
                                        {
                                            myLockAccount.IncrementLockOutCount();
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.LockedOutState))
                                            {
                                                Debug.Print("Failed to set lock state from MainState to LockedOutState");
                                            }
                                        }
                                        else
                                        {
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.InvalidCodeState))
                                            {
                                                Debug.Print("Failed to set lock state from MainState to InvalidCodeState");
                                            }
                                        }

                                        break;
                                    }
                                case LockAccount.eLockCodeType.UserCode:
                                case LockAccount.eLockCodeType.UserCodeLimited:
                                    {
                                        myLockAccount.ResetFailures();

                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.UnlockState))
                                        {
                                            Debug.Print("Failed to set lock state from MainState to UnlockState");
                                        }
                                        break;
                                    }
                                case LockAccount.eLockCodeType.SupervisorCode:
                                    {
                                        myLockAccount.ResetFailures();

                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorMainState))
                                        {
                                            Debug.Print("Failed to set lock state from MainState to SupervisorMainState");
                                        }

                                        break;
                                    }
                                case LockAccount.eLockCodeType.Cancel:
                                    {
                                        //Cancel in main screen causes copyright and version to be displayed
                                        myDispScreen.ShowSignOnScreen();
                                        Thread.Sleep(5000);
                                        //Leave state == MainState
                                        break;
                                    }
                                default:
                                    {
                                        //Any other return cancels operation; leaves state == MainState; 
                                        break;
                                    }
                            }
                            // If lock code valid supervisor code, set state to SupervisorMainState

                            break;
                        }
                    case LockStateMachine.eLockState.InvalidCodeState:
                        {
                            myDispScreen.ShowInvalidCodeScreen();
                            myLockAccount.SaveData();
                            Thread.Sleep(1000);

                            if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                            {
                                Debug.Print("Failed to set lock state from InvalidCodeState to MainState");
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.LockedOutState:
                        {
                            //Display invalid code and locked out
                            myDispScreen.ShowInvalidCodeScreen();
                            myDispScreen.ShowLockedOutScreen();
                            myLockAccount.SaveData();

                            //Sleep locked out time
                            Thread.Sleep(myLockAccount.GetLockOutTime());

                            myLockAccount.ResetFailures();

                            if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                            {
                                Debug.Print("Failed to set lock state from LockedOutState to MainState");
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.UnlockState:
                        {
                            myDispScreen.ShowUnlockScreen();
                            CPLDX0.Write(bUnlock);
                            Thread.Sleep(5000);
                            CPLDX0.Write(bLock);
                            myLockAccount.ResetFailures();

                            if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                            {
                                Debug.Print("Failed to set lock state from UnlockState to MainState");
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorMainState:
                        {
                            byte bCurrentLine = 2;

                            myDispScreen.ShowSupervisorMainScreen();

                            bCurrentLine = myDispScreen.ManageSelectorLine(false, true, true, false, bCurrentLine);

                            if (bCurrentLine == (byte)DispScreen.bCancel)
                            {
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorMainState to MainState");
                                }
                            }
                            else
                            {
                                switch (bCurrentLine)
                                {
                                    case 2:
                                        {
                                            //Set State UnlockState
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.UnlockState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorMainState to UnlockState");
                                            }
                                            break;
                                        }
                                    case 3:
                                        {
                                            //set state SupervisorSelectorState
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorMainState to SupervisorSelectorState");
                                            }
                                            break;
                                        }
                                    default:
                                        {
                                            // Illegal display line
                                            String ErrorString = "Illegal display line in SupervisorMainState: ";
                                            String ErrorLockState = bCurrentLine.ToString();
                                            Debug.Print(String.Concat(ErrorString, ErrorLockState));
                                            //Stay in SupervisorMainState
                                            break;
                                        }
                                }
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorLockCodeState:
                        {
                            myDispScreen.ShowAddRemoveScreen();

                            //Get keypad input
                            uiLockCode = myDispScreen.GetLockCode(bEcho);

                            // If lock code valid user code, set state to UnlockState
                            switch (myLockAccount.CheckLockCode(uiLockCode))
                            {
                                case LockAccount.eLockCodeType.InvalidCode:
                                    {
                                        //Check if lock codes available
                                        if (myLockAccount.GetNumLockCodesAvailable() > 0)
                                        {
                                            // Need to enter lock code entry state
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorAddLockCodeState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorLockCodeState to SupervisorAddLockCodeState");
                                            }
                                        }
                                        else
                                        {
                                            // No lock codes available, so must cancel add operation
                                            myDispScreen.ShowNoAvailableLockCodesScreen();
                                            Thread.Sleep(5000);
                                            //Any other return cancels operation
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorLockCodeState to SupervisorSelectorState");
                                            }
                                        }

                                        break;
                                    }
                                case LockAccount.eLockCodeType.UserCode:
                                case LockAccount.eLockCodeType.UserCodeLimited:
                                    {
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorRemoveUserLockState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorLockCodeState to SupervisorRemoveLockState");
                                        }

                                        break;
                                    }
                                case LockAccount.eLockCodeType.SupervisorCode:
                                    {
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorRemoveSupervisorLockState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorLockCodeState to SupervisorRemoveLockState");
                                        }

                                        break;
                                    }
                                default:
                                    {
                                        //Any other return cancels operation
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorLockCodeState to SupervisorSelectorState");
                                        }

                                        break;
                                    }
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorAddLockCodeState:
                        {
                            byte bCurrentLine = 3;
                            byte bSelectCol = 1;
                            myDispScreen.ShowAddRemoveAddScreen();

                            bSelectCol = myDispScreen.ManageSelectorColumn(1, 7, bCurrentLine);

                            switch (bSelectCol)
                            {
                                case 1:
                                    {
                                        //User account selected
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorAddLockCodeUserState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorAddLockCodeState to SupervisorAddLockCodeUserState");
                                        }
                                        break;
                                    }
                                case 7:
                                    {
                                        //Supervisor account selected
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorAddLockCodeSupervisorState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorAddLockCodeState to SupervisorAddLockCodeSupervisorState");
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        //Operation canceled
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorLockCodeState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorAddLockCodeState to SupervisorLockCodeState");
                                        }
                                        break;
                                    }
                            }

                            break;
                        }
                    case LockStateMachine.eLockState. SupervisorAddLockCodeUserState:
                        {
                            int iLimitedUse;

                            myDispScreen.ShowAddLockCodeUserScreen();

                            //Select Limited use option
                            iLimitedUse = myDispScreen.ManageSpinValueSelect(20);
                            if (iLimitedUse == (int) DispScreen.bCancel)
                            {
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorLockCodeState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to SupervisorLockCodeState");
                                }
                            }
                            else
                            {
                               
                                //Clear selector
                                myDispScreen.ClearSelector(4);
                                //Scroll screen up
                                myDispScreen.ScrollScreenUp();
                                //Prompt to save new lock code
                                myDispScreen.ShowSaveLockCodePrompt();
                                //Get save/cancel
                                byte bResponse = myDispScreen.ManageSelectorColumn(11, 16, 4);
                                if (bResponse == 11)
                                {
                                    //Store new lock code
                                    LockAccount.eLockCodeType eCodeType;
                                    if (iLimitedUse < 0)
                                    {
                                        eCodeType = LockAccount.eLockCodeType.UserCode;
                                    }
                                    else
                                    {
                                        eCodeType = LockAccount.eLockCodeType.UserCodeLimited;
                                    }

                                    //Attempt to store new lock code
                                    if (myLockAccount.SetLock(uiLockCode, eCodeType, iLimitedUse))
                                    {
                                        //Operation completed
                                        myDispScreen.ShowNewLockCodeAdded();
                                        myLockAccount.SaveData();
                                        Thread.Sleep(1000);
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to MainState");
                                        }
                                    }
                                    else
                                    {
                                        //All available lock codes have been used
                                        myDispScreen.ShowNoAvailableLockCodesScreen();
                                        Thread.Sleep(5000);
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to MainState");
                                        }
                                    }
                                }
                                else
                                {
                                    //Cancel
                                    if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorLockCodeState))
                                    {
                                        Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to SupervisorLockCodeState");
                                    }
                                }
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorAddLockCodeSupervisorState:
                        {
                            myDispScreen.ShowAddLockCodeSupervisorScreen();

                            //Scroll screen up
                            myDispScreen.ScrollScreenUp();
                            //Prompt to save new lock code
                            myDispScreen.ShowSaveLockCodePrompt();
                            //Get save/cancel
                            byte bResponse = myDispScreen.ManageSelectorColumn(11, 16, 4);
                            if (bResponse == 11)
                            {
                                //Attempt to store new lock code
                                if (myLockAccount.SetLock(uiLockCode, LockAccount.eLockCodeType.SupervisorCode))
                                {
                                    //Operation completed
                                    myDispScreen.ShowNewLockCodeAdded();
                                    myLockAccount.SaveData();
                                    Thread.Sleep(1000);
                                    if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                                    {
                                        Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to MainState");
                                    }
                                }
                                else
                                {
                                    //All available lock codes have been used
                                    myDispScreen.ShowNoAvailableLockCodesScreen();
                                    Thread.Sleep(5000);
                                    if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                    {
                                        Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to MainState");
                                    }
                                }
                            }
                            else
                            {
                                //Cancel
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorLockCodeState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorAddLockCodeUserState to SupervisorLockCodeState");
                                }
                            }
                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorRemoveUserLockState:
                        {
                            myDispScreen.ShowAddRemoveRemoveUserCodeScreen();

                            //Manage code delete
                            byte bResponse = myDispScreen.ManageSelectorColumn(10, 15, 4);
                            if (bResponse == 10)
                            {
                                myLockAccount.RemoveLockCode(uiLockCode);
                                myDispScreen.ShowConfirmRemovedScreen();
                                myLockAccount.SaveData();
                                Thread.Sleep(1000);
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorRemoveUserLockState to MainState");
                                }
                            }
                            else
                            {
                                //Cancel
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorRemoveUserLockState to SupervisorSelectorState");
                                }
                            }
                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorRemoveSupervisorLockState:
                        {
                            myDispScreen.ShowAddRemoveRemoveSupervisorCodeScreen();

                            //Manage code delete
                            byte bResponse = myDispScreen.ManageSelectorColumn(10, 15, 4);
                            if (bResponse == 10)
                            {
                                myLockAccount.RemoveLockCode(uiLockCode);
                                myDispScreen.ShowConfirmRemovedScreen();
                                myLockAccount.SaveData();
                                Thread.Sleep(1000);
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorRemoveSupervisorLockState to MainState");
                                }
                            }
                            else
                            {
                                //Cancel
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorRemoveSupervisorLockState to SupervisorSelectorState");
                                }
                            }
                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorSelectorState:
                        {
                            byte bCurrentLine = 1;

                            myDispScreen.ShowSuperSelectorScreen();

                            bCurrentLine = myDispScreen.ManageSelectorLine(true, true, true, true, bCurrentLine);

                            if (bCurrentLine == (byte)DispScreen.bCancel)
                            {
                                if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorMainState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorSelectorState to SupervisorMainState");
                                }
                            }
                            else
                            {
                                switch (bCurrentLine)
                                {
                                    case 1:
                                        {
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorLockCodeState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSelectorState to SupervisorLockCodeState");
                                            }
                                            break;
                                        }
                                    case 2:
                                        {
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorViewLogState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSelectorState to SupervisorViewLogState");
                                            }
                                            break;
                                        }
                                    case 3:
                                        {
                                            // Save current security rules
                                            iSaveRetries = myLockAccount.GetMaxCodeFailures();
                                            iSaveLockOut = myLockAccount.GetLockOutMinutes();

                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSetSecurityRulesState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSelectorState to SupervisorSetSecurityRulesState");
                                            }
                                            break;
                                        }
                                    case 4:
                                        {
                                            //Fast exit to main state
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.MainState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSelectorState to MainState");
                                            }
                                            break;
                                        }
                                }
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorSetSecurityRulesState:
                        {
                            byte bCurrentLine = 2;

                            myDispScreen.ShowSecurityRulesScreen(myLockAccount.GetMaxCodeFailures().ToString(), myLockAccount.GetLockOutMinutes().ToString());

                            bCurrentLine = myDispScreen.ManageSelectorLine(false, true, true, true, bCurrentLine);

                            if (bCurrentLine == (byte)DispScreen.bCancel)
                            {
                                //Restore original security rules
                                myLockAccount.SetLockOutTime(iSaveLockOut);
                                myLockAccount.SetMaxCodeFailures((byte)iSaveRetries);
                                myDispScreen.ShowSecurityRulesScreen(myLockAccount.GetMaxCodeFailures().ToString(), myLockAccount.GetLockOutMinutes().ToString());
                                myDispScreen.ShowSecurityRulesCanceled();
                                myLockAccount.SaveData();
                                Thread.Sleep(1000);

                                if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                {
                                    Debug.Print("Failed to set lock state from SupervisorSetSecurityRulesState to SupervisorSelectorState");
                                }
                            }
                            else
                            {
                                switch (bCurrentLine)
                                {
                                    case 2:
                                        {
                                            //Set Maximum Retries State
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSetSecurityRulesRetriesState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSetSecurityRulesState to SupervisorSetSecurityRulesRetriesState");
                                            }
                                            break;
                                        }
                                    case 3:
                                        {
                                            //set lock timeout state
                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSetSecurityRulesLockoutState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSetSecurityRulesState to SupervisorSetSecurityRulesLockoutState");
                                            }
                                            break;
                                        }
                                    case 4:
                                        {
                                            //Show saved indication
                                            myDispScreen.ShowSecurityRulesModified();
                                            myLockAccount.SaveData();
                                            Thread.Sleep(1000);

                                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                            {
                                                Debug.Print("Failed to set lock state from SupervisorSetSecurityRulesState to SupervisorSelectorState");
                                            }
                                            break;
                                        }
                                    default:
                                        {
                                            // Illegal display line
                                            String ErrorString = "Illegal display line in SupervisorSetSecurityRulesState: ";
                                            String ErrorLockState = bCurrentLine.ToString();
                                            Debug.Print(String.Concat(ErrorString, ErrorLockState));
                                            //Stay in SupervisorSetSecurityRulesState
                                            break;
                                        }
                                }
                            }

                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorSetSecurityRulesRetriesState:
                        {
                            int iNewRetries;

                            //Manage maximum retries entry
                            iNewRetries = myDispScreen.ManageSpinValueSelect(9, 1, myLockAccount.GetMaxCodeFailures(), 2, 19);
                            if ((iNewRetries >= 1) && (iNewRetries <= 9))
                            {
                                //Return value in range, so set new value; otherwise cancel set operation
                                myLockAccount.SetMaxCodeFailures((byte) iNewRetries);
                            }

                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSetSecurityRulesState))
                            {
                                Debug.Print("Failed to set lock state from SupervisorSetSecurityRulesRetriesState to SupervisorSetSecurityRulesState");
                            }
                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorSetSecurityRulesLockoutState:
                        {
                            int iNewLockOutMinutes;

                            //Manage lockout minutes entry
                            iNewLockOutMinutes = myDispScreen.ManageSpinValueSelect(60, 1, myLockAccount.GetLockOutMinutes(), 3, 19);
                            if ((iNewLockOutMinutes >= 1) && (iNewLockOutMinutes <= 60))
                            {
                                //Return value in range, so set new value; otherwise cancel set operation
                                myLockAccount.SetLockOutTime(iNewLockOutMinutes);
                            }

                            if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSetSecurityRulesState))
                            {
                                Debug.Print("Failed to set lock state from SupervisorSetSecurityRulesLockoutState to SupervisorSetSecurityRulesState");
                            }
                            break;
                        }
                    case LockStateMachine.eLockState.SupervisorViewLogState:
                        {
                            int iIdx = 0;
                            string[] sScrollArray = new string[myLockAccount.GetMaxLockCodes() + 3];

                            //Add invalid code count
                            sScrollArray[iIdx++] = "Invalid count: " + myLockAccount.GetTotalFailedCodeCount().ToString();
                            //Add lock out count
                            sScrollArray[iIdx++] = "LockOut count: " + myLockAccount.GetLockOutCount().ToString();
                            sScrollArray[iIdx++] = myLockAccount.GetFirstAcctString();
                            while (true)
                            {
                                sScrollArray[iIdx] = myLockAccount.GetNextAcctString();
                                if (string.Equals(sScrollArray[iIdx], ""))
                                {
                                    //Set index to last viewable string
                                    --iIdx;
                                    break;
                                }
                                iIdx++;
                            }
                            switch (myDispScreen.ManageViewLogScreen(sScrollArray, iIdx))
                            {
                                case eLogViewState.DontResetCounts:
                                    {
                                        //Don't clear log counts
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorViewLogState to SupervisorSelectorState");
                                        }
                                        break;
                                    }
                                case eLogViewState.ResetCounts:
                                    {
                                        //Clear log counts
                                        myLockAccount.ClearLogStatistics();

                                        //Show confirm clear log counts
                                        myDispScreen.ShowConfirmLogClear();

                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorViewLogState to SupervisorSelectorState");
                                        }
                                        myLockAccount.SaveData();
                                        Thread.Sleep(1000);
                                        break;
                                    }
                                default:
                                    {
                                        //Cancel
                                        if (!myLockState.SetLockState(LockStateMachine.eLockState.SupervisorSelectorState))
                                        {
                                            Debug.Print("Failed to set lock state from SupervisorViewLogState to SupervisorSelectorState");
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                    default:
                        {
                            // Illegal lock state
                            String ErrorString = "Illegal lock state: ";
                            String ErrorLockState = eCurrentLockState.ToString();
                            Debug.Print(String.Concat(ErrorString, ErrorLockState));
                            break;
                        }
                }
            }
        }
    }

    //Extend the SPI LCD driver to manage the 4-line x 20 character display
    public class SPI_LCD_4LDriver : SPI_LCDDriver
    {
        // Override the line length of the display
        protected static new int ilineLength = 20;

        private const int iLineIndex = 2;
        protected byte[] bSetCursorPreamble = new byte[] { 0xF5, 0x8A };
        protected byte[] bWriteAtCursorPreamble = new byte[] { 0xF5, 0x80 };

        public byte[] LCD4LLineNoWrap(string sDisplayString, byte bLine)
        {
            // Construct the display lines
            //Size the array to manage write preamble, 1 blank char for selector indicator, and actual string
            byte[] bWriteMessageLine = new byte[bWriteLCDLine1Preamble.Length + sDisplayString.Length + 1];

            //Truncate string to size of display
            string sTrimString = "";
            for (int i = 0; (i < ilineLength) && (i < sDisplayString.Length); i++)
            {
                sTrimString += sDisplayString[i];
            }

            switch (bLine)
            {
                case 1:
                {
                    bWriteMessageLine = base.LCDLine1(sTrimString);
                    break;
                }
                case 2:
                {
                    bWriteMessageLine = base.LCDLine2(sTrimString);
                    break;
                }
                case 3:
                case 4:
                {
                    bWriteMessageLine = base.LCDLine1(sTrimString);
                    // Overwrite line index
                    bWriteMessageLine[iLineIndex] = bLine;
                    break;
                }
                default:
                {
                    bWriteMessageLine = base.LCDLine1NoWrap("LCD Line Index Error");
                    break;
                }
            }

            return bWriteMessageLine;
        }

        public byte[] WriteByteAtCursor(byte bChar)
        {
            //Write a single byte to the current cursor location
            byte[] bWriteMsg = new byte[bWriteAtCursorPreamble.Length + 1];
            bWriteAtCursorPreamble.CopyTo(bWriteMsg, 0);
            bWriteMsg[bWriteAtCursorPreamble.Length] = bChar;
            return bWriteMsg;
        }

        public byte[] WriteStringAtCursor(string sMsg)
        {
            // Convert string literals to byte arrays
            byte[] bMsg = new byte[sMsg.Length];

            // convert literal unicode strings to byte arrays
            System.Text.UTF8Encoding Encoding = new System.Text.UTF8Encoding();
            bMsg = Encoding.GetBytes(sMsg);

            byte[] bWriteMsg = new byte[bWriteAtCursorPreamble.Length + sMsg.Length];
            bWriteAtCursorPreamble.CopyTo(bWriteMsg, 0);
            bMsg.CopyTo(bWriteMsg, bWriteAtCursorPreamble.Length);

            return bWriteMsg;
        }

        public byte[] WriteByteChar(byte bChar, byte bDisplayLine, byte bColumn)
        {
            //Write a one-byte character to any line or column - Note: no range checking
            byte[] bWriteSelectorMsg = new byte[bWriteLCDLine1Preamble.Length + 1];
            bWriteLCDLine1Preamble.CopyTo(bWriteSelectorMsg, 0);
            bWriteSelectorMsg[bWriteSelectorMsg.Length - 1] = bChar;
            bWriteSelectorMsg[2] = bDisplayLine;
            bWriteSelectorMsg[3] = bColumn;
            return bWriteSelectorMsg;
        }
        public byte[] WriteByteChar(byte bChar, byte bDisplayLine)
        {
            //Overload: Write a one-byte character to the first column of any line
            byte[] bWriteSelectorMsg = new byte[bWriteLCDLine1Preamble.Length + 1];
            bWriteLCDLine1Preamble.CopyTo(bWriteSelectorMsg, 0);
            bWriteSelectorMsg[bWriteSelectorMsg.Length - 1] = bChar;
            bWriteSelectorMsg[2] = bDisplayLine;
            return bWriteSelectorMsg;
        }

        public byte[] SetCursor(byte bLine, byte bColumn)
        {
            byte[] bWriteCursorMsg = new byte[bSetCursorPreamble.Length + 2];
            bSetCursorPreamble.CopyTo(bWriteCursorMsg, 0);
            bWriteCursorMsg[bSetCursorPreamble.Length] = bLine;
            bWriteCursorMsg[bSetCursorPreamble.Length + 1] = bColumn;
            return bWriteCursorMsg;
        }
    }

    public class DispScreen
    {
        // LCD
        protected static byte bSelector = 0xC7;
        protected static byte bAsterisk = 0x2A;
        protected static byte bExclamation = 0x21;
        protected static byte bQuestion = 0x3F;
        protected static byte bBackSpace = 0xC8;
        protected static byte bSpace = 0x20;

        protected static string sBlankLine = "                    " ;

        protected static string sMainScrn1 = " Enter Lock Code:";

        protected static string sSuperMainScrn1 = " Unlock";
        protected static string sSuperMainScrn2 = " Supervisor Mode";

        protected static string sSuperSelectorScrn1 = " Add/Remove Code";
        protected static string sSuperSelectorScrn2 = " View Log";
        protected static string sSuperSelectorScrn3 = " Set Security Rules";
        protected static string sSuperSelectorScrn4 = " Exit";

        protected static string sAddRemoveScrn1         = " Enter Code";

        protected static string sAddRemoveScrn1Add      = " Adding Code";
        protected static string sAddRemoveScrn3Add      = " User  Supervisor";

        protected static string sAddRemoveScrn3AddUser = " User Code";
        protected static string sAddRemoveScrn4AddUser = " Limited Use";

        protected static string sAddRemoveScrn3AddSupervisor = " Supervisor Code";

        protected static string sSaveLockCodePrompt = "Add Code?  Yes  No";

        protected static string sNewLockCodeAdded = "NEW LOCK CODE ADDED";

        protected static string sFailedScrn1 = "Operation Failed";
        protected static string sNoCodesScrn2 = " All Lock Codes Used";
        protected static string sFailedScrn4 = "    Operation Failed";


        protected static string sAddRemoveScrn3Added    = " ADDED";

        protected static string sAddRemoveScrn1Remove = " Remove Code";
        protected static string sAddRemoveScrn3RemoveUser = " User";
        protected static string sAddRemoveScrn3RemoveSupervisor = " Supervisor";
        protected static string sAddRemoveScrn4Remove   = " Remove?  Yes  No";
        protected static string sAddRemoveScrn4Removed   = " REMOVED";

        protected static string sUnlockedScrn1    = "Unlocked";
        protected static string sUnlockedScrn2    = "    Unlocked";
        protected static string sUnlockedScrn3    = "        Unlocked";
        protected static string sUnlockedScrn4    = "            Unlocked";

        protected static string sInvalidCodeScrn1 = "Invalid Lock Code";
        protected static string sInvalidCodeScrn2 = " Invalid Lock Code";
        protected static string sInvalidCodeScrn3 = "  Invalid Lock Code";
        protected static string sInvalidCodeScrn4 = "   Invalid Lock Code";

        protected static string sSecurityRulesScrn1 = " Set Security Rules";
        protected static string sSecurityRulesScrn2 = " Maximum Retries:";
        protected static string sSecurityRulesScrn3 = " Lockout Minutes:";
        protected static string sSecurityRulesScrn4 = " Done";
        protected static string sSecurityRulesScrn4Updated = " Rules Updated";
        protected static string sSecurityRulesScrn4Canceled = " Change Canceled";

        protected static string sViewLogClearPrompt = "Clear log?  Yes  No";
        protected static string sConfirmLogClear = " LOG CLEARED";

        protected static string sLockedOutScrn = "     Locked Out";

        protected static SPI_LCD_4LDriver mySPILCD;
        protected static SPI.Configuration mySPIPortPortConfig;
        protected static SPI mySPIPort;

        // Keypad
        protected byte[] SPIWriteKeyPadBuf = new byte[3] { 0xF5, 0xD1, 0x00 };
        protected byte[] SPIReadKeyPadBuf = new byte[4];
        public const byte bCursorUp     = 0xC5;
        public const byte bCursorDown   = 0xC6;
        public const byte bCursorLeft   = 0xC8;    //Left arrow
        public const byte bCursorRight  = 0xC7;    //Right arrow
        public const byte bEnter        = 0x0D;
        public const byte bCancel       = 0x18;
        public const byte bMaxNumLockCodeDigits = 9;

        public DispScreen()
        {
            // Constructor will clear the screen
            mySPIPortPortConfig = new SPI.Configuration(Pins.GPIO_NONE, false, 0, 0, false, true, 1, SPI.SPI_module.SPI1);
            mySPIPort = new SPI(mySPIPortPortConfig);
            mySPILCD = new SPI_LCD_4LDriver();

            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCDBacklightOn());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCDHomeCursor());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
            Thread.Sleep(6);

        }

        public UInt32 GetLockCode(Boolean bEcho)
        {
            UInt32 iReturnLockCode = 0;
            byte[] bKeyPad;
            bKeyPad = new byte[bMaxNumLockCodeDigits + 1];
            int idx = 0;

            //Clear keypad of any extra entries
            ClearKey();

            while (true)
            {
                bKeyPad[idx] = ReadKey();
                if ((bKeyPad[idx] >= 0x30) && (bKeyPad[idx] <= 0x39))
                {
                    if (idx < bMaxNumLockCodeDigits)
                    {
                        if (bEcho)
                        {
                            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bKeyPad[idx]));
                        }
                        else
                        {
                            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bAsterisk));
                        }
                        Thread.Sleep(6);
                        
                        //Store actual number not ASCII digit and increment idx
                        bKeyPad[idx] &= 0x0F;
                        idx++;
                    }
                    else
                    {
                        //Trying to add too many digits; don't store; don't increment idx; don't advance cursor
                        mySPIPort.Write(mySPILCD.WriteByteAtCursor(bExclamation));
                        Thread.Sleep(6);
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    }
                }
                else if ((bKeyPad[idx] == (byte)DispScreen.bCancel) || (bKeyPad[idx] == (byte)DispScreen.bEnter))
                {
                    break;
                }
                else if (bKeyPad[idx] == bBackSpace)
                {
                    if (idx > 0)
                    {
                        //First clear any character that might be at the cursor location
                         mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSpace));
                        Thread.Sleep(6);
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                        Thread.Sleep(6);

                        //Backup cursor and idx
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                        Thread.Sleep(6);
                        idx--;

                        //Clear new cursor location
                         mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSpace));
                        Thread.Sleep(6);
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                   }
                }
                else
                {
                    //Trying to add non-numeric character; don't store; don't increment idx; don't advance cursor
                    mySPIPort.Write(mySPILCD.WriteByteAtCursor(bQuestion));
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                }
            }

            if (bKeyPad[idx] == (byte)DispScreen.bEnter)
            {
                idx--;
                for (UInt32 uiPwrTen = 1; idx >= 0; idx--)
                {
                    iReturnLockCode += bKeyPad[idx] * uiPwrTen;
                    uiPwrTen *= 10;
                }
            }

            return iReturnLockCode;
        }

        public byte ManageSelectorLine(Boolean bLine1, Boolean bLine2, Boolean bline3, Boolean bLine4, byte bStartLine)
        {
            // NOTE: Cursor-Down is increasing line number on the display and vice versa
            byte bReturnVal = bCancel;
            byte bKeyPad, bCurrentLine = bStartLine;
            Boolean[] boValidDispLines = { false, bLine1, bLine2, bline3, bLine4, false };

            //Get keypad input and set state to unlocked or SupervisorSelectorState
            while ((bKeyPad = ReadKey()) != (byte)DispScreen.bEnter)
            {
                if (bKeyPad == (byte)DispScreen.bCancel)
                {
                    break;
                }

                if (bKeyPad == (byte)DispScreen.bCursorUp)
                {
                    if (bCurrentLine != 1)
                    {
                        byte idx;
                        for (idx = (byte)(bCurrentLine - 1); idx > 0; idx--)
                        {
                            if (boValidDispLines[idx])
                            {
                                break;
                            }
                        }

                        if (boValidDispLines[idx])
                        {
                            //Move selector down
                            ClearSelector(bCurrentLine);
                            bCurrentLine = idx;
                            WriteSelector(bCurrentLine);
                        }
                    }
                }

                if (bKeyPad == (byte)DispScreen.bCursorDown)
                {
                    if (bCurrentLine != 4)
                    {
                        byte idx;
                        for (idx = (byte)(bCurrentLine + 1); idx < 5; idx++)
                        {
                            if (boValidDispLines[idx])
                            {
                                break;
                            }
                        }

                        if (boValidDispLines[idx])
                        {
                            //Move selector up
                            ClearSelector(bCurrentLine);
                            bCurrentLine = idx;
                            WriteSelector(bCurrentLine);
                        }
                    }
                }

            }

            if (bKeyPad == (byte)DispScreen.bCancel)
            {
                bReturnVal = bKeyPad;
            }
            else
            {
                bReturnVal = bCurrentLine;
            }

            return bReturnVal;
        }

        public int ManageSpinValueSelect(int iMaxVal)
        {
            int iReturnVal = -1;
            byte bKeyPad;

            mySPIPort.Write(mySPILCD.WriteStringAtCursor("Off"));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);

            while ((bKeyPad = ReadKey()) != (byte)DispScreen.bEnter)
            {
                if (bKeyPad == (byte)DispScreen.bCancel)
                {
                    break;
                }

                else if (bKeyPad == (byte)DispScreen.bCursorUp)
                {
                    if (iReturnVal == -1)
                    {
                        iReturnVal = 1;
                    }
                    else if (iReturnVal < iMaxVal)
                    {
                        iReturnVal++;
                    }

                    mySPIPort.Write(mySPILCD.WriteStringAtCursor("   "));
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.WriteStringAtCursor(iReturnVal.ToString()));
                    for (int i = 0; i < iReturnVal.ToString().Length; i++)
                    {
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                        Thread.Sleep(6);
                    }
                }
                else if (bKeyPad == (byte)DispScreen.bCursorDown)
                {
                    if (iReturnVal == 1)
                    {
                        iReturnVal = -1;
                    }
                    else if (iReturnVal > 1)
                    {
                        iReturnVal--;
                    }

                    mySPIPort.Write(mySPILCD.WriteStringAtCursor("   "));
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    Thread.Sleep(6);
                    mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    Thread.Sleep(6);
                    if (iReturnVal == -1)
                    {
                        mySPIPort.Write(mySPILCD.WriteStringAtCursor("Off"));
                        Thread.Sleep(6);
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                        Thread.Sleep(6);
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                        Thread.Sleep(6);
                        mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                    }
                    else
                    {
                        mySPIPort.Write(mySPILCD.WriteStringAtCursor(iReturnVal.ToString()));
                        for (int i = 0; i < iReturnVal.ToString().Length; i++)
                        {
                            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                            Thread.Sleep(6);
                        }
                    }
                }
            }
            // No conflict between possible column values and Cancel value
            if (bKeyPad == (byte)DispScreen.bCancel)
            {
                iReturnVal = (int) bKeyPad;
            }

            return iReturnVal;
        }

        public int ManageSpinValueSelect(int iMaxVal, int iMinVal, int iCurrentVal, byte bLine, byte bCol)
        {
            int iReturnVal = iCurrentVal;
            int iMaxClearChars = (int) iMaxVal.ToString().Length;
            byte bKeyPad;

            mySPIPort.Write(mySPILCD.SetCursor(bLine, bCol));
            mySPIPort.Write(mySPILCD.WriteStringAtCursor(iCurrentVal.ToString()));
            Thread.Sleep(6);
            for (int i = 0; i < iCurrentVal.ToString().Length; i++)
            {
                mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                Thread.Sleep(6);
            }

            while ((bKeyPad = ReadKey()) != (byte)DispScreen.bEnter)
            {
                if (bKeyPad == (byte)DispScreen.bCancel)
                {
                    break;
                }

                else if (bKeyPad == (byte)DispScreen.bCursorUp)
                {
                    if (iReturnVal < iMaxVal)
                    {
                        iReturnVal++;

                        for (int i = 0; i < iMaxClearChars; i++)
                        {
                            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSpace));
                            Thread.Sleep(6);
                        }
                        for (int i = 0; i < iMaxClearChars; i++)
                        {
                            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                            Thread.Sleep(6);
                        }
                        mySPIPort.Write(mySPILCD.WriteStringAtCursor(iReturnVal.ToString()));
                        for (int i = 0; i < iReturnVal.ToString().Length; i++)
                        {
                            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                            Thread.Sleep(6);
                        }
                    }
                }
                else if (bKeyPad == (byte)DispScreen.bCursorDown)
                {
                    if (iReturnVal > iMinVal)
                    {
                        iReturnVal--;

                        for (int i = 0; i < iMaxClearChars; i++)
                        {
                            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSpace));
                            Thread.Sleep(6);
                        }
                        for (int i = 0; i < iMaxClearChars; i++)
                        {
                            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                            Thread.Sleep(6);
                        }
                        mySPIPort.Write(mySPILCD.WriteStringAtCursor(iReturnVal.ToString()));
                        for (int i = 0; i < iReturnVal.ToString().Length; i++)
                        {
                            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
                            Thread.Sleep(6);
                        }
                    }
                }
            }
            // For no conflict between possible return values and Cancel value, return a number out of range for Cancel
            if (bKeyPad == (byte)DispScreen.bCancel)
            {
                iReturnVal = (int)(iMaxVal + 1);
            }

            return iReturnVal;
        }

        public byte ManageSelectorColumn(byte bCol1, byte bCol2, byte bLine)
        {
            byte bReturnVal = bCancel;
            byte bKeyPad, bCurrentColumn = bCol1;
            byte[] bValidCursorColumns = { bCol1, bCol2 };
            byte bIdx = 0;
            const byte bIdxMax = 1;

            //Write selector to first valid column
            WriteSelector(bLine, bValidCursorColumns[bIdx]);

            while ((bKeyPad = ReadKey()) != (byte)DispScreen.bEnter)
            {
                if (bKeyPad == (byte)DispScreen.bCancel)
                {
                    break;
                }

                else if (bKeyPad == (byte)DispScreen.bCursorLeft)
                {
                    if (bIdx > 0)
                    {
                        ClearSelector(bLine, bValidCursorColumns[bIdx]);
                        bIdx--;
                        WriteSelector(bLine, bValidCursorColumns[bIdx]);
                    }
                }

                else if (bKeyPad == (byte)DispScreen.bCursorRight)
                {
                    if (bIdx < bIdxMax)
                    {
                        ClearSelector(bLine, bValidCursorColumns[bIdx]);
                        bIdx++;
                        WriteSelector(bLine, bValidCursorColumns[bIdx]);
                    }
                }

            }

            // No conflict between possible column values and Cancel value
            if (bKeyPad == (byte)DispScreen.bCancel)
            {
                bReturnVal = bKeyPad;
            }
            else
            {
                bReturnVal = bValidCursorColumns[bIdx];
            }

            return bReturnVal;
        }

        public void WriteSelector(byte bLine)
        {
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, bLine));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void WriteSelector(byte bLine, byte bCol)
        {
            mySPIPort.Write(mySPILCD.SetCursor(bLine, bCol));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSelector));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ClearSelector(byte bLine, byte bCol)
        {
            mySPIPort.Write(mySPILCD.SetCursor(bLine, bCol));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSpace));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ClearSelector(byte bLine)
        {
            mySPIPort.Write(mySPILCD.WriteByteChar(bSpace, bLine));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ScrollScreenUp()
        {
            mySPIPort.Write(mySPILCD.LCDShiftDisplayUp());
        }

        public void ScrollScreenDown()
        {
            mySPIPort.Write(mySPILCD.LCDShiftDisplayDown());
        }

        public void ShowMainScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sMainScrn1, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowSupervisorMainScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSuperMainScrn1, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSuperMainScrn2, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowSuperSelectorScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSuperSelectorScrn1, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSuperSelectorScrn2, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSuperSelectorScrn3, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSuperSelectorScrn4, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowAddRemoveScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn1, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowAddRemoveRemoveUserCodeScreen()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn1Remove, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSpace, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn3RemoveUser, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn4Remove, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 4, 10));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowAddRemoveRemoveSupervisorCodeScreen()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn1Remove, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSpace, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn3RemoveSupervisor, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn4Remove, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 4, 10));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowConfirmRemovedScreen()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn4Removed, 4));
        }

        public void ShowAddRemoveAddScreen()
        {
            mySPIPort.Write(mySPILCD.WriteByteChar(bSpace, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn1Add, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn3Add, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowAddLockCodeUserScreen()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn3AddUser, 3));
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn4AddUser, 4));
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.SetCursor(4, 14));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowAddLockCodeSupervisorScreen()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sAddRemoveScrn3AddSupervisor, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowSaveLockCodePrompt()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSaveLockCodePrompt, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.SetCursor(4, 11));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteByteAtCursor(bSelector));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowNewLockCodeAdded()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sNewLockCodeAdded, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowNoAvailableLockCodesScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sFailedScrn1, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sNoCodesScrn2, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowUnlockScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sUnlockedScrn1, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sUnlockedScrn2, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sUnlockedScrn3, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sUnlockedScrn4, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowSignOnScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(Resources.GetString(Resources.StringResources.String3), 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(Resources.GetString(Resources.StringResources.String4), 3));
        }

        public void ShowInvalidCodeScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sInvalidCodeScrn1, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sInvalidCodeScrn2, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sInvalidCodeScrn3, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sInvalidCodeScrn4, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowLockedOutScreen()
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sInvalidCodeScrn3, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sLockedOutScrn, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sLockedOutScrn, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sInvalidCodeScrn3, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowConfirmLogClear()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sConfirmLogClear, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public App.eLogViewState ManageViewLogScreen(string[] sScrollArray, int iMaxIndex)
        {
            App.eLogViewState eReturnVal = App.eLogViewState.CancelLogView;
            int iIdx = 0;
            byte bKeyPad;

            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());

            for (byte i = 1; ((i <= 3) && (iIdx <= iMaxIndex)); i++)
            {
                mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sScrollArray[iIdx++], i));
                Thread.Sleep(6);
            }
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sViewLogClearPrompt, 4));
            Thread.Sleep(6);

            //Get keypad input and set state to unlocked or SupervisorSelectorState
            while ((bKeyPad = ReadKey()) != (byte)DispScreen.bEnter)
            {
                if (bKeyPad == (byte)DispScreen.bCancel)
                {
                    break;
                }

                else if (bKeyPad == (byte)DispScreen.bCursorUp)
                {
                    //Cursor up scrolls screen down
                    if ((iIdx - 3) > 0)
                    {
                        //Can scroll down
                        mySPIPort.Write(mySPILCD.LCDShiftDisplayDown());
                        mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sViewLogClearPrompt, 4));
                        mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sScrollArray[iIdx - 4], 1));
                        iIdx--;
                        Thread.Sleep(6);

                    }
                }

                else if (bKeyPad == (byte)DispScreen.bCursorDown)
                {
                    //Cursor down scrolls screen up
                    if (iIdx <= iMaxIndex)
                    {
                        //Can Scroll up
                        mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
                        mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sScrollArray[iIdx], 4));
                        mySPIPort.Write(mySPILCD.LCDShiftDisplayUp());
                        mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sViewLogClearPrompt, 4));
                        iIdx++;
                    }
                }
            }
            if (bKeyPad == bEnter)
            {
                //Manage clear
                byte bSelect = ManageSelectorColumn(12, 17, 4);
                if (bSelect == 12)
                {
                    eReturnVal = App.eLogViewState.ResetCounts;
                }
                else if (bSelect == 17)
                {
                    eReturnVal = App.eLogViewState.DontResetCounts;
                }
                else
                {
                    eReturnVal = App.eLogViewState.CancelLogView;
                }
            }
            else
            {
                eReturnVal = App.eLogViewState.CancelLogView;
            }

            return eReturnVal;
        }

        public void ShowSecurityRulesScreen(string sMaxRetries, string sLockoutMin)
        {
            mySPIPort.Write(mySPILCD.LCDClear());
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSecurityRulesScrn1, 1));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSecurityRulesScrn2, 2));
            Thread.Sleep(6);

            //Show current values
            mySPIPort.Write(mySPILCD.SetCursor(2, 19));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteStringAtCursor(sMaxRetries));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.SetCursor(3, 19));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.WriteStringAtCursor(sLockoutMin));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSecurityRulesScrn3, 3));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSecurityRulesScrn4, 4));
            Thread.Sleep(6);

            mySPIPort.Write(mySPILCD.WriteByteChar(bSelector, 2));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDMoveCursorLeft());
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOn());
        }

        public void ShowSecurityRulesCanceled()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSecurityRulesScrn4Canceled, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ShowSecurityRulesModified()
        {
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sBlankLine, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCD4LLineNoWrap(sSecurityRulesScrn4Updated, 4));
            Thread.Sleep(6);
            mySPIPort.Write(mySPILCD.LCDBlinkblockCursorOff());
        }

        public void ClearKey()
        {
            byte bKeyChar = 0x00;

            do
            {
                mySPIPort.WriteRead(SPIWriteKeyPadBuf, SPIReadKeyPadBuf);
                Thread.Sleep(6);
            } while (bKeyChar != 0x00);
        }

        public byte ReadKey()
        {
            byte bKeyChar = 0x00;

            do
            {
                mySPIPort.WriteRead(SPIWriteKeyPadBuf, SPIReadKeyPadBuf);

                //Check for a valid keypad character
                if ((SPIReadKeyPadBuf[3] >= 0x61) && (SPIReadKeyPadBuf[3] <= 0x70))
                {
                    switch (SPIReadKeyPadBuf[3])
                    {
                        case 0x61:
                            {
                                bKeyChar = 0x31;
                                break;
                            }
                        case 0x62:
                            {
                                bKeyChar = 0x32;
                                break;
                            }
                        case 0x63:
                            {
                                bKeyChar = 0x33;
                                break;
                            }
                        case 0x64:
                            {
                                bKeyChar = 0x0D;    //Enter
                                break;
                            }
                        case 0x65:
                            {
                                bKeyChar = 0x34;
                                break;
                            }
                        case 0x66:
                            {
                                bKeyChar = 0x35;
                                break;
                            }
                        case 0x67:
                            {
                                bKeyChar = 0x36;
                                break;
                            }
                        case 0x68:
                            {
                                bKeyChar = 0xC5;    //Up arrow
                                break;
                            }
                        case 0x69:
                            {
                                bKeyChar = 0x37;
                                break;
                            }
                        case 0x6A:
                            {
                                bKeyChar = 0x38;
                                break;
                            }
                        case 0x6B:
                            {
                                bKeyChar = 0x39;
                                break;
                            }
                        case 0x6C:
                            {
                                bKeyChar = 0xC6;    //Down arrow
                                break;
                            }
                        case 0x6D:
                            {
                                bKeyChar = 0x18;    //Cancel
                                break;
                            }
                        case 0x6E:
                            {
                                bKeyChar = 0x30;
                                break;
                            }
                        case 0x6F:
                            {
                                bKeyChar = 0xC8;    //Left arrow
                                break;
                            }
                        case 0x70:
                            {
                                bKeyChar = 0xC7;    //Right arrow
                                break;
                            }
                        default:
                            {
                                bKeyChar = 0x00;
                                break;
                            }
                    }
                }

                if (bKeyChar == 0x00)
                {
                    //If no char, delay 1/4 second before next read
                    Thread.Sleep(250);
                }

            } while (bKeyChar == 0x00);

            return bKeyChar;
        }
    }

    public class LockStateMachine
    {
        public enum eLockState
        {
            NullState,
            MainState,
            UnlockState,
            InvalidCodeState,
            LockedOutState,
            SupervisorMainState,
            SupervisorSelectorState,
            SupervisorLockCodeState,
            SupervisorAddLockCodeState,
            SupervisorAddLockCodeUserState,
            SupervisorAddLockCodeSupervisorState,
            SupervisorRemoveUserLockState,
            SupervisorRemoveSupervisorLockState,
            SupervisorSetSecurityRulesState,
            SupervisorSetSecurityRulesRetriesState,
            SupervisorSetSecurityRulesLockoutState,
            SupervisorViewLogState,
            TotalStates
        }

        private static eLockState LockState;

        public LockStateMachine()
        {
            LockState = eLockState.MainState;
        }

        public eLockState GetLockState()
        {
            return LockState;
        }

        public Boolean SetLockState(eLockState SetState)
        {
            Boolean bReturnVal = false;

            if (SetState < eLockState.TotalStates)
            {
                LockState = SetState;
                bReturnVal = true;
            }

            return bReturnVal;
        }
    }

    public class LockAccount
    {
        // Start EWR declarations
        // Extended Weak Reference declarations for non-volatile storage
        private static ExtendedWeakReference myAppData;

        private static class TypeAppDataFlag { }

        [Serializable]
        private sealed class theNonVolatileData
        {
            public byte bTotalFailedCount;
            public byte bLockOutCount;
            public byte bMaxAllowedFailures;
            public Int32 iLockOutTime;
            public ValidLockCode[] ValidLockCodeArray = new ValidLockCode[iMaximumLockCodes];

            public void SetTotalFailedCount(byte bNewTotalFailedCount)
            {
                bTotalFailedCount = bNewTotalFailedCount;
            }

            public void SetLockOutCount(byte bNewLockOutCount)
            {
                bLockOutCount = bNewLockOutCount;
            }

            public void SetMaxAllowedFailures(byte bNewMaxAllowedFailures)
            {
                bMaxAllowedFailures = bNewMaxAllowedFailures;
            }

            public void SetLockOutTime(Int32 iNewLockOutTime)
            {
                iLockOutTime = iNewLockOutTime;
            }

            public Int32 GetLockOutTime()
            {
                return iLockOutTime;
            }

            public void InitializeValidLockCodeArray()
            {
                //Create objects for valid code array
                for (int i = 0; i < iMaximumLockCodes; i++)
                {
                    ValidLockCodeArray[i] = new ValidLockCode();
                }
                //Create default user and supervisor accounts
                ValidLockCodeArray[0].SetLockCode(123);
                ValidLockCodeArray[0].SetLockCodeType(eLockCodeType.UserCode);

                ValidLockCodeArray[1].SetLockCode(321);
                ValidLockCodeArray[1].SetLockCodeType(eLockCodeType.SupervisorCode);
            }
            public void IncrementTotalFailedCount()
            {
                bTotalFailedCount++;
            }

            public byte GetTotalFailedCount()
            {
                return bTotalFailedCount;
            }

            public void IncrememtLockOutCount()
            {
                bLockOutCount++;
            }

            public byte GetLockOutCount()
            {
                return bLockOutCount;
            }

            public byte GetMaxAllowedFailures()
            {
                return bMaxAllowedFailures;
            }
        }

        private static theNonVolatileData myNonVolatileData;
        // End EWR declarations

        private static byte bFailedCount;
        private static int iGetFirstNextIdx = new int();
        private const int iMaximumLockCodes = 32;

        [Serializable]
        protected class ValidLockCode
        {
            private UInt32 uiLockCode;
            private eLockCodeType eCodeType;
            private int iLimitedUse;
            private UInt64 uiUsageCount;


            public ValidLockCode()
            {
                uiLockCode = 0;
                eCodeType = eLockCodeType.InvalidCode;
                iLimitedUse = -1;
                uiUsageCount = 0;
            }

            public void SetLockCode(UInt32 uiNewLockCode)
            {
                uiLockCode = uiNewLockCode;
            }

            public  UInt32 GetLockCode()
            {
                return uiLockCode;
            }

            public void SetLockCodeType(eLockCodeType eNewCodeType)
            {
                eCodeType = eNewCodeType;
            }

            public eLockCodeType GetLockCodeType()
            {
                return eCodeType;
            }

            public void SetLimitedUse(int iNewLimit)
            {
                iLimitedUse = iNewLimit;
            }

            public int GetLimitedUse()
            {
                return iLimitedUse;
            }

            public void DecrementLimitedUse()
            {
                iLimitedUse--;
            }

            public void IncrementUsageCount()
            {
                uiUsageCount++;
            }

            public void ClearUsageCount()
            {
                uiUsageCount = 0;
            }

            public UInt64 GetUsageCount()
            {
                return uiUsageCount;
            }
        }

        public enum eLockCodeType
        {
            UserCode,
            UserCodeLimited,
            SupervisorCode,
            InvalidCode,
            Cancel,
            NumCodeTypes
        }

        public LockAccount()
        {
            // Do lock count initialization
            bFailedCount = 0;
        }

        //Extended Weak Reference methods
        public Boolean GetData()
        {
            Boolean boolReturnVal = true;

            myAppData = ExtendedWeakReference.RecoverOrCreate(typeof(TypeAppDataFlag), 0, ExtendedWeakReference.c_SurvivePowerdown);
            myAppData.Priority = (Int32)ExtendedWeakReference.PriorityLevel.Critical;

            myNonVolatileData = myAppData.Target as theNonVolatileData;

            if (myNonVolatileData == null)
            {
                Debug.Print("EWR non-volatile data: first boot or lost");

                myNonVolatileData = new theNonVolatileData();
                myNonVolatileData.SetTotalFailedCount(0);
                myNonVolatileData.SetLockOutCount(0);
                myNonVolatileData.SetMaxAllowedFailures(3);
                myNonVolatileData.SetLockOutTime(1000 * 60 * 5);
                myNonVolatileData.InitializeValidLockCodeArray();

                boolReturnVal = false;
            }

            return boolReturnVal;
        }

        public void SaveData()
        {
            myAppData.Target = myNonVolatileData;
            Thread.Sleep(4000);
        }

        //End Extended Weak Reference methods

        public Boolean SetLock(UInt32 uiLockCode, eLockCodeType LockCodeType)
        {
            Boolean bReturnVal = false;

            //Assumes either UserCode or SupervisorCode, set use unlimited
            for (int i = 0; i < iMaximumLockCodes; i++)
            {
                if (myNonVolatileData.ValidLockCodeArray[i].GetLockCodeType() == eLockCodeType.InvalidCode)
                {
                    //Lock code element is available to be used
                    myNonVolatileData.ValidLockCodeArray[i].SetLockCode(uiLockCode);
                    myNonVolatileData.ValidLockCodeArray[i].SetLockCodeType(LockCodeType);
                    bReturnVal = true;
                    break;
                }
            }

            return bReturnVal;
        }

        public Boolean SetLock(UInt32 uiLockCode, eLockCodeType LockCodeType, int iUseCount)
        {
            Boolean bReturnVal = false;

            //Assumes limit use lock code
            for (int i = 0; i < iMaximumLockCodes; i++)
            {
                if (myNonVolatileData.ValidLockCodeArray[i].GetLockCodeType() == eLockCodeType.InvalidCode)
                {
                    //Lock code element is available to be used
                    myNonVolatileData.ValidLockCodeArray[i].SetLockCode(uiLockCode);
                    myNonVolatileData.ValidLockCodeArray[i].SetLockCodeType(LockCodeType);
                    myNonVolatileData.ValidLockCodeArray[i].SetLimitedUse(iUseCount);
                    bReturnVal = true;
                    break;
                }
            }

            return bReturnVal;
        }

        public void RemoveLockCode(UInt32 uiLockCode)
        {
            for (int i = 0; i < iMaximumLockCodes; i++)
            {
                if (myNonVolatileData.ValidLockCodeArray[i].GetLockCode() == uiLockCode)
                {
                    myNonVolatileData.ValidLockCodeArray[i].SetLockCodeType(eLockCodeType.InvalidCode);
                    myNonVolatileData.ValidLockCodeArray[i].SetLockCode(0);
                    myNonVolatileData.ValidLockCodeArray[i].SetLimitedUse(-1);
                    myNonVolatileData.ValidLockCodeArray[i].ClearUsageCount();
                    break;
                }
            }
        }

        public int GetNumLockCodesAvailable()
        {
            int iReturnVal = 0;

            for (int i = 0; i < iMaximumLockCodes; i++)
            {
                if (myNonVolatileData.ValidLockCodeArray[i].GetLockCodeType() == eLockCodeType.InvalidCode)
                {
                    iReturnVal++;
                }
            }

            return iReturnVal;
        }


        public eLockCodeType CheckLockCode(UInt32 uiLockCode)
        {
            eLockCodeType ReturnVal = eLockCodeType.InvalidCode;
            eLockCodeType LockCodeType;
            UInt32 LockCode;

            if (uiLockCode == 0)
            {
                ReturnVal = eLockCodeType.Cancel;
            }
            else
            {
                for (int i = 0; i < iMaximumLockCodes; i++)
                {
                    LockCodeType = myNonVolatileData.ValidLockCodeArray[i].GetLockCodeType();
                    if (LockCodeType != eLockCodeType.InvalidCode)
                    {
                        LockCode = myNonVolatileData.ValidLockCodeArray[i].GetLockCode();
                        if (LockCode == uiLockCode)
                        {
                            //Check for limited use account
                            if (LockCodeType == eLockCodeType.UserCodeLimited)
                            {
                                if (myNonVolatileData.ValidLockCodeArray[i].GetLimitedUse() == 0)
                                {
                                    //Expire the limited use lock code
                                    myNonVolatileData.ValidLockCodeArray[i].SetLockCodeType(eLockCodeType.InvalidCode);
                                    myNonVolatileData.ValidLockCodeArray[i].ClearUsageCount();
                                    LockCodeType = eLockCodeType.InvalidCode;
                                }
                                else
                                {
                                    //Decrement limited use count
                                    myNonVolatileData.ValidLockCodeArray[i].DecrementLimitedUse();
                                    //Update lock code access count
                                    myNonVolatileData.ValidLockCodeArray[i].IncrementUsageCount();
                                }
                            }
                            else
                            {
                                //Update lock code access count
                                myNonVolatileData.ValidLockCodeArray[i].IncrementUsageCount();
                            }

                            ReturnVal = LockCodeType;
                            break;
                        }
                    }
                }
            }

            return ReturnVal;
        }

        public void IncrementLockOutCount()
        {
            myNonVolatileData.IncrememtLockOutCount();
        }

        public void IncrementFailedCodeCount()
        {
            bFailedCount++;
            myNonVolatileData.IncrementTotalFailedCount();
        }

        public byte GetLockOutCount()
        {
            return myNonVolatileData.GetLockOutCount();
        }

        public int GetMaxLockCodes()
        {
            return iMaximumLockCodes;
        }

        public byte GetTotalFailedCodeCount()
        {
            return myNonVolatileData.GetTotalFailedCount();
        }

        public void SetMaxCodeFailures(byte bMaxFailures)
        {
            myNonVolatileData.SetMaxAllowedFailures(bMaxFailures);
        }

        public void ClearLogStatistics()
        {
            myNonVolatileData.SetTotalFailedCount(0);
            myNonVolatileData.SetLockOutCount(0);

            for (int i = 0; i < iMaximumLockCodes; i++)
            {
                myNonVolatileData.ValidLockCodeArray[i].ClearUsageCount();
            }
        }

        public byte GetMaxCodeFailures()
        {
            return myNonVolatileData.GetMaxAllowedFailures();
        }

        public Boolean CheckCodeFailures()
        {
            Boolean bReturnVal = true;

            if (bFailedCount < myNonVolatileData.GetMaxAllowedFailures())
            {
                bReturnVal = false;
            }
            return bReturnVal;
        }

        public void ResetFailures()
        {
            bFailedCount = 0;
        }

        public Int32 GetLockOutTime()
        {
            // mSec
            return myNonVolatileData.GetLockOutTime();
        }

        public Int32 GetLockOutMinutes()
        {
            return (Int32) ((myNonVolatileData.GetLockOutTime() / 60) / 1000);
        }

        public void SetLockOutTime(Int32 iNewLockOutMinutes)
        {
            myNonVolatileData.SetLockOutTime(1000 * 60 * iNewLockOutMinutes);
        }

        public string GetFirstAcctString()
        {
            string sReturnString = "";
            eLockCodeType eLockCode;

            for (iGetFirstNextIdx = 0; iGetFirstNextIdx < iMaximumLockCodes; iGetFirstNextIdx++)
            {
                eLockCode = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCodeType();
                if (eLockCode != eLockCodeType.Cancel)
                {
                    switch (eLockCode)
                    {
                        case eLockCodeType.SupervisorCode:
                            {
                                sReturnString = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCode().ToString() + " S: " + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetUsageCount().ToString();
                                break;
                            }
                        case eLockCodeType.UserCode:
                            {
                                sReturnString = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCode().ToString() + " U: " + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetUsageCount().ToString();
                                break;
                            }
                        case eLockCodeType.UserCodeLimited:
                            {
                                sReturnString = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCode().ToString() + " UL" + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLimitedUse() + ": " + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetUsageCount().ToString();
                                break;
                            }
                    }

                    if (eLockCode != eLockCodeType.InvalidCode)
                    {
                        iGetFirstNextIdx++;
                        break;
                    }
                }
                else
                {
                    iGetFirstNextIdx++;
                    sReturnString = "";
                    break;
                }
            }

            return sReturnString;
        }

        public string GetNextAcctString()
        {
            string sReturnString = "";
            eLockCodeType eLockCode;

            for ( ; iGetFirstNextIdx < iMaximumLockCodes; iGetFirstNextIdx++)
            {
                eLockCode = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCodeType();
                if (eLockCode != eLockCodeType.Cancel)
                {
                    switch (eLockCode)
                    {
                        case eLockCodeType.SupervisorCode:
                            {
                                sReturnString = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCode().ToString() + " S: " + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetUsageCount().ToString();
                                break;
                            }
                        case eLockCodeType.UserCode:
                            {
                                sReturnString = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCode().ToString() + " U: " + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetUsageCount().ToString();
                                break;
                            }
                        case eLockCodeType.UserCodeLimited:
                            {
                                sReturnString = myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLockCode().ToString() + " UL" + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetLimitedUse() + ": " + myNonVolatileData.ValidLockCodeArray[iGetFirstNextIdx].GetUsageCount().ToString();
                                break;
                            }
                    }

                    if (eLockCode != eLockCodeType.InvalidCode)
                    {
                        iGetFirstNextIdx++;
                        break;
                    }
                }
                else
                {
                    iGetFirstNextIdx++;
                    sReturnString = "";
                    break;
                }
            }

            return sReturnString;
        }
    }
}
